
#pragma once

#if defined (ENABLE_DEBUG_OUTPUT)
#include <stdio.h>
#endif

//=============================================================================
//
//	BitBoard9 class
//  
//	9x9 bit board.
//	The bit board is implemented with 4 64bit unsigned ints, each representing 
//	the 8x8 board from (0,0) to (7,7); from (3,0) to (10,7); from (0,3) to (7,10);
//	from (3,3) to (10,10). 	(Note that coordinates 0 and 10 represents the edge coordinates)
//	The overlaps exist to facilitate the retrieval of the surroundings of a given point.
//
//=============================================================================

// definitions to translate points to usable indeces
#define INDEX_(x,y) ((x+1) >> 3) + (((y+1) >> 3) << 1)
#define INDEX(pt) INDEX_(pt.x, pt.y)
#define BIT_RANK_(x,y,index) ((x - ((index & 1) | ((index << 1) & 2))) | ((y - ((index & 2) | (index >> 1))) << 3))
#define BIT_RANK(pt, index)  BIT_RANK_(pt.x, pt.y, index)
#define BIT_(x,y,index) 1ui64 << BIT_RANK_(x, y, index) 
#define BIT(pt,index) BIT_(pt.x, pt.y, index)

// point on the board
typedef struct Point
{
	unsigned char x;
	unsigned char y;

	bool operator ==(Point other)
	{
		return x == other.x && y == other.y;
	}

	Point() : x(0), y(0) {}
	Point(unsigned char xPos, unsigned char yPos) : x(xPos), y(yPos) {}
} Point;

typedef unsigned char StoneColor;

static const StoneColor BLACK_STONECOLOR = 0;
static const StoneColor WHITE_STONECOLOR = 1;

// move, i.e., point with color
typedef struct Move
{
	StoneColor color;
	Point point;

	Move() : color(0xff), point(0,0) {}
	Move(StoneColor col, Point pt) : color(col), point(pt) {}
	Move(StoneColor col, unsigned char x, unsigned char y) : color(col), point(x,y) {}
} Move;

class BitBoard9
{
public:
	// constructor
	BitBoard9() { Clear(); }

	// Clears the board 
	void Clear()
	{
		stones[0] = 0;
		stones[1] = 0;
		stones[2] = 0;
		stones[3] = 0;
	}

	// Fill the edge stones
	void FillEdge()
	{
		stones[0] |= 0x01010101010101ffui64;
		stones[1] |= 0x80808080808080ffui64;
		stones[2] |= 0xff01010101010101ui64;
		stones[3] |= 0xff80808080808080ui64;
	}

	// Set the neighbors of a point
	// precondition: the board is empty
	void InitializeWithPointsAround(Point pt)
	{
		int index = INDEX(pt);
		unsigned long long bit = BIT(pt,index);
		
		stones[index] = (bit >> 1) | (bit << 1) | (bit >> 8) | (bit << 8);
		SetAllOverlappedStones();
	}

	// Set a stone (bit) at a point
	void SetStone(Point pt)
	{
		int index = INDEX(pt);
		stones[index] |= BIT(pt,index);
	}

	// Remove a stone from a point
	void RemoveStone(Point pt)
	{
		int index = INDEX(pt);
		stones[index] ^= BIT(pt,index);
	}

	// Return true if the board does not have a stone at the point
	bool IsEmpty(Point pt)
	{
		int index = INDEX(pt);
		return (stones[index] & BIT(pt,index) ) == 0;
	}

	// Return true if the board does not have a stone at (x,y)
	bool IsEmpty(unsigned char x, unsigned char y)
	{
		int index = INDEX_(x,y);
		return (stones[index] & BIT_(x,y,index) ) == 0;
	}

	// sets the common empty points of the given two boards around the point
	// precondition: the board is empty
	void SetLibertiesAround(Point pt, BitBoard9& board1, BitBoard9& board2)
	{
		int index = INDEX(pt);
		unsigned long long bit = BIT(pt,index);
		
		stones[index] = (~board1.stones[index]) & (~board2.stones[index]) & ((bit >> 1) | (bit << 1) | (bit >> 8) | (bit << 8));
		SetAllOverlappedStones();
	}

	// sets the common empty points of the two boards, masked by the BitBoard9 neighbors 
	void SetLibertiesOf(BitBoard9& neighbors, BitBoard9& board1, BitBoard9& board2)
	{
		stones[0] = neighbors.stones[0] & (~board1.stones[0]) & (~board2.stones[0]);
		stones[1] = neighbors.stones[1] & (~board1.stones[1]) & (~board2.stones[1]);
		stones[2] = neighbors.stones[2] & (~board1.stones[2]) & (~board2.stones[2]);
		stones[3] = neighbors.stones[3] & (~board1.stones[3]) & (~board2.stones[3]);
	}

	// sets the common stones of the neighbots and the given board
	void SetSurroundingStonesOf(BitBoard9& neighbors, BitBoard9& board)
	{
		stones[0] = neighbors.stones[0] & board.stones[0];
		stones[1] = neighbors.stones[1] & board.stones[1];
		stones[2] = neighbors.stones[2] & board.stones[2];
		stones[3] = neighbors.stones[3] & board.stones[3];
	}

	// sets the common stones of the neighbor and the sum of the two boards
	void SetSurroundingStonesOf(BitBoard9& neighbors, BitBoard9& board1, BitBoard9& board2)
	{
		stones[0] = neighbors.stones[0] & (board1.stones[0] | board2.stones[0]);
		stones[1] = neighbors.stones[1] & (board1.stones[1] | board2.stones[1]);
		stones[2] = neighbors.stones[2] & (board1.stones[2] | board2.stones[2]);
		stones[3] = neighbors.stones[3] & (board1.stones[3] | board2.stones[3]);
	}
	
	// takes one stone from the given board, and sets this board to be a board with exactly that one stone
	void CopyOneStoneFrom(BitBoard9& target)
	{
		stones[0] = target.stones[0] & ((~target.stones[0]) + 1);
		stones[1] = target.stones[1] & ((~target.stones[1]) + (target.stones[0] == 0));
		stones[2] = target.stones[2] & ((~target.stones[2]) + ((target.stones[0] | target.stones[1]) == 0));
		stones[3] = target.stones[3] & ((~target.stones[3]) + ((target.stones[0] | target.stones[1] | target.stones[2]) == 0));

		SetAllOverlappedStones();
	}
	
	// cuts out one stone from the given board, and sets this board to be a board with exactly that one stone
	void CutOneStoneFrom(BitBoard9& target)
	{
		stones[0] = target.stones[0] & ((~target.stones[0]) + 1);
		stones[1] = target.stones[1] & ((~target.stones[1]) + (target.stones[0] == 0));
		stones[2] = target.stones[2] & ((~target.stones[2]) + ((target.stones[0] | target.stones[1]) == 0));
		stones[3] = target.stones[3] & ((~target.stones[3]) + ((target.stones[0] | target.stones[1] | target.stones[2]) == 0));

		SetAllOverlappedStones();

		target.RemoveStones(*this);
	}
	
	// takes in a 3x3 pattern on a 8x4 32bit int board, and see if 
	// that pattern matches the board at the point, with (2,2) of the pattern moved to the given point
	bool Match3x3Pattern(unsigned int pattern, Point pt)
	{
		int index = INDEX(pt);
		int rank  = BIT_RANK(pt,index);

		// shifts the pattern to rank 9 lower (x-1 and y-1) than the rank of the given point
		unsigned long long expandedPattern = ((unsigned long long)pattern) << (rank - 9);

		return (stones[index] & expandedPattern) == expandedPattern;
	}

	// returns the only point left on the board
	// precondition: there are only one stone left on the board
	void GetTheOnlyPoint(Point* pt)
	{
		// copy the board
		unsigned long long temp0 = stones[0] & 0x007e7e7e7e7e7e00ui64; // get the 6x6 board from (1,1) to (6,6)
		unsigned long long temp1 = stones[1] & 0x0070707070707000ui64; // get the 3x6 board from (7,1) to (9,6)
		unsigned long long temp2 = stones[2] & 0x007e7e7e00000000ui64; // get the 6x3 board from (1,7) to (6,9)
		unsigned long long temp3 = stones[3] & 0x0070707000000000ui64; // get the 3x3 board from (7,7) to (9,9)

		// compress all the x axis info to one line
		unsigned long long tempX = ((temp0           & 0x0000000000007e00ui64) >>  8) |
								   ((temp0           & 0x00000000007e0000ui64) >> 16) |
								   ((temp0           & 0x000000007e000000ui64) >> 24) |
								   (((temp0 | temp2) & 0x0000007e00000000ui64) >> 32) |
								   (((temp0 | temp2) & 0x00007e0000000000ui64) >> 40) |
								   (((temp0 | temp2) & 0x007e000000000000ui64) >> 48) |
								   ((temp1           & 0x0000000000007e00ui64) >>  5) |
								   ((temp1           & 0x00000000007e0000ui64) >> 13) |
								   ((temp1           & 0x000000007e000000ui64) >> 21) |
								   (((temp1 | temp3) & 0x0000007e00000000ui64) >> 29) |
								   (((temp1 | temp3) & 0x00007e0000000000ui64) >> 37) |
								   (((temp1 | temp3) & 0x007e000000000000ui64) >> 45) ;

		// compress all the y axis into two lines
		unsigned long long tempY1 = ((temp0           & 0x0002020202020200ui64) >>  8) |
								    ((temp0           & 0x0004040404040400ui64) >>  9) |
								    ((temp0           & 0x0008080808080800ui64) >> 10) |
								    (((temp0 | temp1) & 0x0010101010101000ui64) >> 11) |
								    (((temp0 | temp1) & 0x0020202020202000ui64) >> 12) |
								    (((temp0 | temp1) & 0x0040404040404000ui64) >> 13) ;

		unsigned long long tempY2 = ((temp2           & 0x0002020200000000ui64) >>  8) |
								    ((temp2           & 0x0004040400000000ui64) >>  9) |
								    ((temp2           & 0x0008080800000000ui64) >> 10) |
								    (((temp2 | temp3) & 0x0010101000000000ui64) >> 11) |
								    (((temp2 | temp3) & 0x0020202000000000ui64) >> 12) |
								    (((temp2 | temp3) & 0x0040404000000000ui64) >> 13) ;

		pt->x = 1 + (tempX > 0x0002) + (tempX > 0x0004) + (tempX > 0x0008) + (tempX > 0x0010) + (tempX > 0x0020) + (tempX > 0x0040) + (tempX > 0x0080) + (tempX > 0x0100);
		pt->y = (tempY1 >= 0x0000000000000002ui64) + (tempY1 >= 0x0000000000000200ui64) + (tempY1 >= 0x0000000000002000ui64) +
				(tempY1 >= 0x0000000002000000ui64) + (tempY1 >= 0x0000000200000000ui64) + (tempY1 >= 0x0000020000000000ui64) +
				((tempY2 > 0) * 6) +
				(tempY2 >= 0x0000000002000000ui64) + (tempY2 >= 0x0000000200000000ui64) + (tempY2 >= 0x0000020000000000ui64) ;
	}

	// returns the number of stones on the board (i.e., the number of valid bits standing
	int CountStones()
	{
		// copy the board
		unsigned long long temp0 = stones[0] & 0x007e7e7e7e7e7e00ui64; // get the 6x6 board from (1,1) to (6,6)
		unsigned long long temp1 = stones[1] & 0x0070707070707000ui64; // get the 3x6 board from (7,1) to (9,6)
		unsigned long long temp2 = stones[2] & 0x007e7e7e00000000ui64; // get the 6x3 board from (1,7) to (6,9)
		unsigned long long temp3 = stones[3] & 0x0070707000000000ui64; // get the 3x3 board from (7,7) to (9,9)

		// rbg𐔂 (Divide & Conquer)
		temp0 = (temp0 & 0x5555555555555555ui64) + ((temp0 & 0xaaaaaaaaaaaaaaaaui64) >> 1);
		temp0 = (temp0 & 0x3333333333333333ui64) + ((temp0 & 0xccccccccccccccccui64) >> 2);
		temp0 = (temp0 & 0x0f0f0f0f0f0f0f0fui64) + ((temp0 & 0xf0f0f0f0f0f0f0f0ui64) >> 4);
		temp0 = (temp0 & 0x00ff00ff00ff00ffui64) + ((temp0 & 0xff00ff00ff00ff00ui64) >> 8);
		temp0 = (temp0 & 0x0000ffff0000ffffui64) + ((temp0 & 0xffff0000ffff0000ui64) >> 16);
		temp0 = (temp0 & 0x00000000ffffffffui64) + ((temp0 & 0xffffffff00000000ui64) >> 32);
		temp1 = (temp1 & 0x5555555555555555ui64) + ((temp1 & 0xaaaaaaaaaaaaaaaaui64) >> 1);
		temp1 = (temp1 & 0x3333333333333333ui64) + ((temp1 & 0xccccccccccccccccui64) >> 2);
		temp1 = (temp1 & 0x0f0f0f0f0f0f0f0fui64) + ((temp1 & 0xf0f0f0f0f0f0f0f0ui64) >> 4);
		temp1 = (temp1 & 0x00ff00ff00ff00ffui64) + ((temp1 & 0xff00ff00ff00ff00ui64) >> 8);
		temp1 = (temp1 & 0x0000ffff0000ffffui64) + ((temp1 & 0xffff0000ffff0000ui64) >> 16);
		temp1 = (temp1 & 0x00000000ffffffffui64) + ((temp1 & 0xffffffff00000000ui64) >> 32);
		temp2 = (temp2 & 0x5555555555555555ui64) + ((temp2 & 0xaaaaaaaaaaaaaaaaui64) >> 1);
		temp2 = (temp2 & 0x3333333333333333ui64) + ((temp2 & 0xccccccccccccccccui64) >> 2);
		temp2 = (temp2 & 0x0f0f0f0f0f0f0f0fui64) + ((temp2 & 0xf0f0f0f0f0f0f0f0ui64) >> 4);
		temp2 = (temp2 & 0x00ff00ff00ff00ffui64) + ((temp2 & 0xff00ff00ff00ff00ui64) >> 8);
		temp2 = (temp2 & 0x0000ffff0000ffffui64) + ((temp2 & 0xffff0000ffff0000ui64) >> 16);
		temp2 = (temp2 & 0x00000000ffffffffui64) + ((temp2 & 0xffffffff00000000ui64) >> 32);
		temp3 = (temp3 & 0x5555555555555555ui64) + ((temp3 & 0xaaaaaaaaaaaaaaaaui64) >> 1);
		temp3 = (temp3 & 0x3333333333333333ui64) + ((temp3 & 0xccccccccccccccccui64) >> 2);
		temp3 = (temp3 & 0x0f0f0f0f0f0f0f0fui64) + ((temp3 & 0xf0f0f0f0f0f0f0f0ui64) >> 4);
		temp3 = (temp3 & 0x00ff00ff00ff00ffui64) + ((temp3 & 0xff00ff00ff00ff00ui64) >> 8);
		temp3 = (temp3 & 0x0000ffff0000ffffui64) + ((temp3 & 0xffff0000ffff0000ui64) >> 16);
		temp3 = (temp3 & 0x00000000ffffffffui64) + ((temp3 & 0xffffffff00000000ui64) >> 32);

		return (int)(temp0 + temp1 + temp2 + temp3);
	}

	// rotate the board 90 degrees clockwise
	void Rotate90Clockwise()
	{
		unsigned long long temp = stones[0];

		stones[0] = stones[2];
		stones[2] = stones[3];
		stones[3] = stones[1];
		stones[1] = temp;

		stones[0] = ((stones[0] & 0xf0f0f0f000000000ui64) >>  4) | 
			        ((stones[0] & 0x0f0f0f0f00000000ui64) >> 32) | 
				    ((stones[0] & 0x00000000f0f0f0f0ui64) << 32) | 
					((stones[0] & 0x000000000f0f0f0fui64) <<  4);

		stones[0] = ((stones[0] & 0xcccc0000cccc0000ui64) >>  2) | 
			        ((stones[0] & 0x3333000033330000ui64) >> 16) | 
				    ((stones[0] & 0x0000cccc0000ccccui64) << 16) | 
					((stones[0] & 0x0000333300003333ui64) <<  2);

		stones[0] = ((stones[0] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[0] & 0x5500550055005500ui64) >> 8) | 
				    ((stones[0] & 0x00aa00aa00aa00aaui64) << 8) | 
					((stones[0] & 0x0055005500550055ui64) << 1);

		stones[1] = ((stones[1] & 0xf0f0f0f000000000ui64) >>  4) | 
			        ((stones[1] & 0x0f0f0f0f00000000ui64) >> 32) | 
				    ((stones[1] & 0x00000000f0f0f0f0ui64) << 32) | 
					((stones[1] & 0x000000000f0f0f0fui64) <<  4);

		stones[1] = ((stones[1] & 0xcccc0000cccc0000ui64) >>  2) | 
			        ((stones[1] & 0x3333000033330000ui64) >> 16) | 
				    ((stones[1] & 0x0000cccc0000ccccui64) << 16) | 
					((stones[1] & 0x0000333300003333ui64) <<  2);

		stones[1] = ((stones[1] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[1] & 0x5500550055005500ui64) >> 8) | 
				    ((stones[1] & 0x00aa00aa00aa00aaui64) << 8) | 
					((stones[1] & 0x0055005500550055ui64) << 1);

		stones[2] = ((stones[2] & 0xf0f0f0f000000000ui64) >>  4) | 
			        ((stones[2] & 0x0f0f0f0f00000000ui64) >> 32) | 
				    ((stones[2] & 0x00000000f0f0f0f0ui64) << 32) | 
					((stones[2] & 0x000000000f0f0f0fui64) <<  4);

		stones[2] = ((stones[2] & 0xcccc0000cccc0000ui64) >>  2) | 
			        ((stones[2] & 0x3333000033330000ui64) >> 16) | 
				    ((stones[2] & 0x0000cccc0000ccccui64) << 16) | 
					((stones[2] & 0x0000333300003333ui64) <<  2);

		stones[2] = ((stones[2] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[2] & 0x5500550055005500ui64) >> 8) | 
				    ((stones[2] & 0x00aa00aa00aa00aaui64) << 8) | 
					((stones[2] & 0x0055005500550055ui64) << 1);

		stones[3] = ((stones[3] & 0xf0f0f0f000000000ui64) >>  4) | 
			        ((stones[3] & 0x0f0f0f0f00000000ui64) >> 32) | 
				    ((stones[3] & 0x00000000f0f0f0f0ui64) << 32) | 
					((stones[3] & 0x000000000f0f0f0fui64) <<  4);

		stones[3] = ((stones[3] & 0xcccc0000cccc0000ui64) >>  2) | 
			        ((stones[3] & 0x3333000033330000ui64) >> 16) | 
				    ((stones[3] & 0x0000cccc0000ccccui64) << 16) | 
					((stones[3] & 0x0000333300003333ui64) <<  2);

		stones[3] = ((stones[3] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[3] & 0x5500550055005500ui64) >> 8) | 
				    ((stones[3] & 0x00aa00aa00aa00aaui64) << 8) | 
					((stones[3] & 0x0055005500550055ui64) << 1);
	}

	// filp the board horizonttaly (x,y) -> (10-x. y)
	void FlipHorizontally()
	{
		unsigned long long temp = stones[0];
		stones[0] = stones[1];
		stones[1] = temp;

		temp = stones[3];
		stones[3] = stones[2];
		stones[2] = temp;
		

		stones[0] = ((stones[0] & 0xf0f0f0f000000000ui64) >> 4) | 
			        ((stones[0] & 0x0f0f0f0f00000000ui64) << 4) | 
				    ((stones[0] & 0x00000000f0f0f0f0ui64) >> 4) | 
					((stones[0] & 0x000000000f0f0f0fui64) << 4);

		stones[0] = ((stones[0] & 0xcccc0000cccc0000ui64) >> 2) | 
			        ((stones[0] & 0x3333000033330000ui64) << 2) | 
				    ((stones[0] & 0x0000cccc0000ccccui64) >> 2) | 
					((stones[0] & 0x0000333300003333ui64) << 2);

		stones[0] = ((stones[0] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[0] & 0x5500550055005500ui64) << 1) | 
				    ((stones[0] & 0x00aa00aa00aa00aaui64) >> 1) | 
					((stones[0] & 0x0055005500550055ui64) << 1);

		stones[1] = ((stones[1] & 0xf0f0f0f000000000ui64) >> 4) | 
			        ((stones[1] & 0x0f0f0f0f00000000ui64) << 4) | 
				    ((stones[1] & 0x00000000f0f0f0f0ui64) >> 4) | 
					((stones[1] & 0x000000000f0f0f0fui64) << 4);

		stones[1] = ((stones[1] & 0xcccc0000cccc0000ui64) >> 2) | 
			        ((stones[1] & 0x3333000033330000ui64) << 2) | 
				    ((stones[1] & 0x0000cccc0000ccccui64) >> 2) | 
					((stones[1] & 0x0000333300003333ui64) << 2);

		stones[1] = ((stones[1] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[1] & 0x5500550055005500ui64) << 1) | 
				    ((stones[1] & 0x00aa00aa00aa00aaui64) >> 1) | 
					((stones[1] & 0x0055005500550055ui64) << 1);

		stones[2] = ((stones[2] & 0xf0f0f0f000000000ui64) >> 4) | 
			        ((stones[2] & 0x0f0f0f0f00000000ui64) << 4) | 
				    ((stones[2] & 0x00000000f0f0f0f0ui64) >> 4) | 
					((stones[2] & 0x000000000f0f0f0fui64) << 4);

		stones[2] = ((stones[2] & 0xcccc0000cccc0000ui64) >> 2) | 
			        ((stones[2] & 0x3333000033330000ui64) << 2) | 
				    ((stones[2] & 0x0000cccc0000ccccui64) >> 2) | 
					((stones[2] & 0x0000333300003333ui64) << 2);

		stones[2] = ((stones[2] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[2] & 0x5500550055005500ui64) << 1) | 
				    ((stones[2] & 0x00aa00aa00aa00aaui64) >> 1) | 
					((stones[2] & 0x0055005500550055ui64) << 1);

		stones[3] = ((stones[3] & 0xf0f0f0f000000000ui64) >> 4) | 
			        ((stones[3] & 0x0f0f0f0f00000000ui64) << 4) | 
				    ((stones[3] & 0x00000000f0f0f0f0ui64) >> 4) | 
					((stones[3] & 0x000000000f0f0f0fui64) << 4);

		stones[3] = ((stones[3] & 0xcccc0000cccc0000ui64) >> 2) | 
			        ((stones[3] & 0x3333000033330000ui64) << 2) | 
				    ((stones[3] & 0x0000cccc0000ccccui64) >> 2) | 
					((stones[3] & 0x0000333300003333ui64) << 2);

		stones[3] = ((stones[3] & 0xaa00aa00aa00aa00ui64) >> 1) | 
			        ((stones[3] & 0x5500550055005500ui64) << 1) | 
				    ((stones[3] & 0x00aa00aa00aa00aaui64) >> 1) | 
					((stones[3] & 0x0055005500550055ui64) << 1);
	}

	// sets all stones on the given board
	void SetStones(BitBoard9& stoneBoard)
	{
		stones[0] |= stoneBoard.stones[0];
		stones[1] |= stoneBoard.stones[1];
		stones[2] |= stoneBoard.stones[2];
		stones[3] |= stoneBoard.stones[3];
	}

	// removes all stones on the given board
	void RemoveStones(BitBoard9& stoneBoard)
	{
		stones[0] ^= stoneBoard.stones[0];
		stones[1] ^= stoneBoard.stones[1];
		stones[2] ^= stoneBoard.stones[2];
		stones[3] ^= stoneBoard.stones[3];
	}

	// update the overlapped parts
	void Update()
	{ 
 		stones[0] =  (stones[0] & 0x007f7f7f7f7f7f7fui64)        | ((stones[1] & 0x1010101010101010ui64) <<  3) | ((stones[2] & 0x000000ff00000000ui64) << 24) | ((stones[3] & 0x0000001000000000ui64) << 27);
		stones[1] = ((stones[0] & 0xf8f8f8f8f8f8f8f8ui64) >>  3) |  (stones[1] & 0xe0e0e0e0e0e0e0e0ui64)        | ((stones[3] & 0x000000ff00000000ui64) << 24);
		stones[2] = ((stones[0] & 0xffffffffff000000ui64) >> 24) |  (stones[2] & 0xffffff0000000000ui64)        | ((stones[3] & 0x1010101010101010ui64) <<  3);
		stones[3] = ((stones[1] & 0xffffffffff000000ui64) >> 24) | ((stones[2] & 0xf8f8f8f8f8f8f8f8ui64) >>  3) |  (stones[3] & 0xe0e0e00000000000ui64);
	}

	// returns true if the board is completely empty
	bool IsEmptyBoard()
	{
		return (0 == (stones[0] | stones[1] | stones[2] | stones[3]));
	}

	// return true if there is at least one stones in common with the given board
	bool HasStonesInCommon(BitBoard9 other)
	{
		return (0 != (stones[0] & other.stones[0])) ||
			   (0 != (stones[1] & other.stones[1])) ||
			   (0 != (stones[2] & other.stones[2])) ||
			   (0 != (stones[3] & other.stones[3])) ;
	}

	// returns true if all stones on the board is on the first line
	bool IsEdgeLevelRank()
	{
		return (stones[0] == (stones[0] & 0x020202020202ff00ui64)) &&
			   (stones[1] == (stones[1] & 0x404040404040ff00ui64)) &&
			   (stones[2] == (stones[2] & 0x00ff020202020202ui64)) &&
			   (stones[3] == (stones[3] & 0x00ff404040404040ui64)) ;
	}

	bool operator ==(BitBoard9 &other) { return stones[0] == other.stones[0] && stones[1] == other.stones[1] && stones[2] == other.stones[2] && stones[3] == other.stones[3]; }
	void operator |=(BitBoard9 &other)
	{
		stones[0] |= other.stones[0];
		stones[1] |= other.stones[1];
		stones[2] |= other.stones[2];
		stones[3] |= other.stones[3];
	}		
	void operator &=(BitBoard9 &other)
	{
		stones[0] &= other.stones[0];
		stones[1] &= other.stones[1];
		stones[2] &= other.stones[2];
		stones[3] &= other.stones[3];
	}		
	void operator -=(BitBoard9 &other)
	{
		stones[0] &= ~other.stones[0];
		stones[1] &= ~other.stones[1];
		stones[2] &= ~other.stones[2];
		stones[3] &= ~other.stones[3];
	}	

#if defined (ENABLE_DEBUG_OUTPUT)
	void Print()
	{
		printf("%02x %02x %02x %02x\n", (int)((stones[0]      ) & 0xff), (int)((stones[1]      ) & 0xff), (int)((stones[2]      ) & 0xff), (int)((stones[3]      ) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >>  8) & 0xff), (int)((stones[1] >>  8) & 0xff), (int)((stones[2] >>  8) & 0xff), (int)((stones[3] >>  8) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >> 16) & 0xff), (int)((stones[1] >> 16) & 0xff), (int)((stones[2] >> 16) & 0xff), (int)((stones[3] >> 16) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >> 24) & 0xff), (int)((stones[1] >> 24) & 0xff), (int)((stones[2] >> 24) & 0xff), (int)((stones[3] >> 24) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >> 32) & 0xff), (int)((stones[1] >> 32) & 0xff), (int)((stones[2] >> 32) & 0xff), (int)((stones[3] >> 32) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >> 40) & 0xff), (int)((stones[1] >> 40) & 0xff), (int)((stones[2] >> 40) & 0xff), (int)((stones[3] >> 40) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >> 48) & 0xff), (int)((stones[1] >> 48) & 0xff), (int)((stones[2] >> 48) & 0xff), (int)((stones[3] >> 48) & 0xff));
		printf("%02x %02x %02x %02x\n", (int)((stones[0] >> 56) & 0xff), (int)((stones[1] >> 56) & 0xff), (int)((stones[2] >> 56) & 0xff), (int)((stones[3] >> 56) & 0xff));
	}
#endif

private:
	// force update the overlapped part (this used mainly for initializing funcs)
	void SetAllOverlappedStones()
	{
		stones[0] |= ((stones[1] & 0x1f1f1f1f1f1f1f1fui64) <<  3) | ((stones[2] & 0x000000ffffffffffui64) << 24) | ((stones[3] & 0x0000001f1f1f1f1f1fui64) << 27);
		stones[1] |= ((stones[0] & 0xf8f8f8f8f8f8f8f8ui64) >>  3) | ((stones[3] & 0x000000ffffffffffui64) << 24);
		stones[2] |= ((stones[0] & 0xffffffffff000000ui64) >> 24) | ((stones[3] & 0x1f1f1f1f1f1f1f1fui64) <<  3);
		stones[3] |= ((stones[1] & 0xffffffffff000000ui64) >> 24) | ((stones[2] & 0xf8f8f8f8f8f8f8f8ui64) >>  3);
	}


private:
	unsigned long long stones[4];
};