
#include "BoardManager.h"

BoardManager::BoardManager()
{
	Clear();
}

void BoardManager::Clear()
{
	currentBoard_        = boardStack_ + 1;		// skip the very first board, because the ko-checking function looks at the board prior to the current board
	currentPrisonerData_ = prisonerDataStack_;

	currentBoard_->Initialize();

	// set up the initial prisoners
	currentPrisonerData_->prisoners_[0] = 0;
	currentPrisonerData_->prisoners_[1] = 0;
}

void BoardManager::SetStone(Move move)
{
	// copy the current board
	*(currentBoard_ + 1) = *(currentBoard_++);

	// copy the current prisoner data
	*(currentPrisonerData_ + 1) = *(currentPrisonerData_++);

	// 
	currentBoard_->SetStone(move);
}

bool BoardManager::RemoveIfCaptured(StoneChain* chain)
{
	if (currentBoard_->IsSurrounded(chain->GetNeighbors()))
	{
		currentBoard_->RemoveStones(chain->GetColor(), chain->GetStones());
		currentPrisonerData_->prisoners_[chain->GetColor()] += chain->GetNumStones();

		return true;
	}

	return false;
}

bool BoardManager::IsEmpty(Point point)
{
	return currentBoard_->IsEmpty(point);
}

bool BoardManager::IsEmpty(unsigned char x, unsigned char y)
{
	return currentBoard_->IsEmpty(x,y);
}

bool BoardManager::IsEmpty(BitBoard9 mask)
{
	return currentBoard_->IsEmpty(mask);
}

void BoardManager::GetLibertiesAround(BitBoard9& libBoard, Point point)
{
	currentBoard_->GetLibertiesAround(libBoard, point);
}

bool BoardManager::IsKoViolated()
{
	return *(currentBoard_) == *(currentBoard_ - 2); 
}

bool BoardManager::IsCaptured(StoneChain* chain)
{
	return currentBoard_->IsSurrounded(chain->GetNeighbors());
}

bool BoardManager::HasStone(Point pt, StoneColor color)
{
	return currentBoard_->HasStone(pt, color);
}

void BoardManager::GetLiberties(StoneChain* chain, BitBoard9* libs)
{
	return currentBoard_->GetLiberties(chain->GetNeighbors(), libs);
}

void BoardManager::GetSurroundingStones(StoneChain* chain, BitBoard9* surrounding)
{
	// retrieve data from the board of opposite color of the chain
	return currentBoard_->GetSurroundingStones(chain->GetNeighbors(), surrounding, chain->GetColor() ^ 1);
}

void BoardManager::GetSurroundingStones(Point pt, BitBoard9* surrounding)
{
	BitBoard9 neighbors;
	neighbors.InitializeWithPointsAround(pt);
	return currentBoard_->GetSurroundingStones(neighbors, surrounding);
}

int BoardManager::GetNumPrisoners(StoneColor color)
{
	return currentPrisonerData_->prisoners_[color];
}

bool BoardManager::Match3x3Pattern(unsigned int pattern, Move move)
{
	return currentBoard_->Match3x3Pattern(pattern, move);
}

bool BoardManager::HasBothColorStoneAround(Point pt)
{
	return currentBoard_->HasBothColorStoneAround(pt);
}

void BoardManager::Undo()
{
	--currentBoard_;
	--currentPrisonerData_;
}

#if defined (ENABLE_DEBUG_OUTPUT)
void BoardManager::ShowBoard()
{
	Point pt;

	// print board
	for (pt.y = 9; pt.y >= 1; pt.y--)
	{
		for (pt.x = 1; pt.x <= 9; pt.x++)
		{
			if (currentBoard_->IsBlack(pt) && currentBoard_->IsWhite(pt))
				printf("? ");
			else if (currentBoard_->IsBlack(pt))
				printf("# ");
			else if (currentBoard_->IsWhite(pt))
				printf("O ");
			else
				printf("+ ");
		}
		printf("\n");
	}

	// print prisoner info
	printf("Black prisoners = %d\n", currentPrisonerData_->prisoners_[0]);
	printf("White prisoners = %d\n", currentPrisonerData_->prisoners_[1]);
}

void BoardManager::ShowBitBoard()
{
	currentBoard_->ShowBitBoard();
}

void BoardManager::ShowChain(StoneChain* chain)
{
	Point pt;

	BitBoard9 libs;
	GetLiberties(chain, &libs);

	printf("===:::===\n");

	printf("Stones = %d, Neighbors = %d, Libs = %d\n", chain->GetNumStones(), chain->GetNumNeighbors(), libs.CountStones());

	for (pt.y = 9; pt.y >= 1; pt.y--)
	{
		for (pt.x = 1; pt.x <= 9; pt.x++)
		{
			if (!chain->GetStones().IsEmpty(pt))		// chain stones
			{
				if (chain->GetColor() == 0)
					printf("# ");
				else
					printf("O ");
			}
			else if (!chain->GetNeighbors().IsEmpty(pt))	// neighbors
			{
				if (currentBoard_->IsEmpty(pt))
					printf("x ");
				else
				{
					printf("* ");
				}
			}
			else
				printf("+ ");
		}
		printf("\n");
	}
}
#endif