
#include "JosekiDictionary.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

JosekiDictionary::JosekiDictionary(const char *josekiFileName)
{
	currentNode_ = rootNode_ = new JosekiNode();
	currentRotation_ = ROTATED_0;

	// read from sgf file
	FILE* josekiFile = fopen(josekiFileName, "r");
	if (josekiFile)
	{
		// skip the first two info lines
		char ch;
		int count = 0;
		while (count < 2)
		{
			fread(&ch, 1, 1, josekiFile);
			if (ch == '\n')
				count++;
		}

		BuildTree(currentNode_, josekiFile, 0);

		fclose(josekiFile);
	}
}

JosekiDictionary::~JosekiDictionary()
{
	delete rootNode_;

	rootNode_ = 0;
	currentNode_ = 0;
}

void JosekiDictionary::Initialize()
{
	currentNode_ = rootNode_;
}

void JosekiDictionary::Proceed(Board currentBoard)
{
	if (!currentNode_)
		return;

	for (int i = 0; i < currentNode_->numChildren_; i++)
	{
		for (unsigned char state = ROTATED_0; state < NUM_ROATATION_STATES; state++)
		{
			if (RotateBoard(currentNode_->children_[i]->boardBeforeMove_, state) == currentBoard)
			{
				currentNode_ = currentNode_->children_[i];
				currentRotation_ = state;
				return;
			}
		}
	}

	// no matching node found, void the tree
	currentNode_ = 0;
}

Point JosekiDictionary::GetNextMove()
{
	if (currentNode_ && currentNode_->numChildren_ > 0)
	{
		return RotatePoint(currentNode_->children_[rand() % currentNode_->numChildren_]->move_, currentRotation_);
	}
	else
	{
		Point passPt;
		passPt.x = 0;
		passPt.y = 0;

		return passPt;
	}
}

void JosekiDictionary::BuildTree(JosekiNode* root, FILE* josekiFile, StoneColor startColor)
{
	char ch;
	Point pt;
	JosekiNode* node = root;
	JosekiNode* newNode;
	Move move;
	move.color = startColor;

	fread(&ch, 1, 1, josekiFile);
	while (ch != ')')
	{
		if (ch == '(')
		{
			// build tree for new variation
			BuildTree(node, josekiFile, move.color);
		}
		else
		{
			assert(ch == ';');

			fread(&ch, 1, 1, josekiFile);	// read and trash color
			fread(&ch, 1, 1, josekiFile);	// read and trash open bracket

			fread(&ch, 1, 1, josekiFile);	// read x coordinate
			pt.x = ch - 'a' + 1;
			fread(&ch, 1, 1, josekiFile);	// read y coordinate
			pt.y = ch - 'a' + 1;

			newNode = new JosekiNode();
			newNode->move_   = pt;
			newNode->parent_ = node;
			newNode->boardBeforeMove_ = node->boardBeforeMove_;

			move.point = pt;
			newNode->boardBeforeMove_.SetStone(move);
			move.color ^= 0x01;

			node->children_[node->numChildren_++] = newNode;
			node = newNode;
			
			fread(&ch, 1, 1, josekiFile);	// read and trash close bracket
		}

		// read char for next iteration
		fread(&ch, 1, 1, josekiFile);
		if (ch == '\n')
		{
			// skip new line character
			fread(&ch, 1, 1, josekiFile);
		}
	}
}

Point JosekiDictionary::RotatePoint(Point pt, unsigned char rotation)
{
	switch (rotation)
	{
	case ROTATED_0:
		return pt;
	case ROTATED_90:
		return Point(10 - pt.y, pt.x);
	case ROTATED_180:
		return Point(10 - pt.x, 10 - pt.y);
	case ROTATED_270:
		return Point(pt.y, 10 - pt.x);
	case MIRROR_ROTATED_0:
		return Point(10 - pt.x, pt.y);
	case MIRROR_ROTATED_90:
		return Point(10 - pt.y, 10 - pt.x);
	case MIRROR_ROTATED_180:
		return Point(pt.x, 10 - pt.y);
	case MIRROR_ROTATED_270:
		return Point(pt.y, pt.x);
	}

	return Point(0,0);
}

Board JosekiDictionary::RotateBoard(Board board, unsigned char rotation)
{
	Board rotatedBoard;

	switch (rotation)
	{
	case ROTATED_0:
		rotatedBoard = board;
		break;
	case ROTATED_90:
		rotatedBoard = board;
		rotatedBoard.Rotate90Clockwise();
		break;
	case ROTATED_180:
		rotatedBoard = board;
		rotatedBoard.Rotate90Clockwise();
		rotatedBoard.Rotate90Clockwise();
		break;
	case ROTATED_270:
		rotatedBoard = board;
		rotatedBoard.Rotate90Clockwise();
		rotatedBoard.Rotate90Clockwise();
		rotatedBoard.Rotate90Clockwise();
		break;
	case MIRROR_ROTATED_0:
		rotatedBoard = board;
		rotatedBoard.FlipHorizontally();
		break;
	case MIRROR_ROTATED_90:
		rotatedBoard = board;
		rotatedBoard.FlipHorizontally();
		rotatedBoard.Rotate90Clockwise();
		break;
	case MIRROR_ROTATED_180:
		rotatedBoard = board;
		rotatedBoard.FlipHorizontally();
		rotatedBoard.Rotate90Clockwise();
		rotatedBoard.Rotate90Clockwise();
		break;
	case MIRROR_ROTATED_270:
		rotatedBoard = board;
		rotatedBoard.FlipHorizontally();
		rotatedBoard.Rotate90Clockwise();
		rotatedBoard.Rotate90Clockwise();
		rotatedBoard.Rotate90Clockwise();
		break;
	}

	return rotatedBoard;
}

JosekiDictionary::JosekiNode::JosekiNode()
{
	numChildren_ = 0;
	parent_ = 0;

	move_.x = 0;
	move_.y = 0;
}

JosekiDictionary::JosekiNode::~JosekiNode()
{
	for (int i = 0; i < numChildren_; i++)
	{
		delete children_[i];
	}
}