
#include "StoneChainManager.h"

#include "BoardStonePointEnumerator.h"

#include <new>

//#include <string.h>

#define BOARD_INDEX_(x,y) ((x) + 11*(y))
#define BOARD_INDEX(pt)  BOARD_INDEX_(pt.x, pt.y)

#define BOARD_INDEX_NORTH(pt) BOARD_INDEX_(pt.x, pt.y - 1)
#define BOARD_INDEX_SOUTH(pt) BOARD_INDEX_(pt.x, pt.y + 1)
#define BOARD_INDEX_WEST(pt)  BOARD_INDEX_(pt.x - 1, pt.y)
#define BOARD_INDEX_EAST(pt)  BOARD_INDEX_(pt.x + 1, pt.y)

StoneChainManager::StoneChainManager()
{
	// init stone chains
	stoneChains_ = (StoneChain*)malloc(sizeof(StoneChain) * MAX_STONE_CHAINS);
	for (int i = 0; i < MAX_STONE_CHAINS; i++)
	{
		// replacement new the StoneChains with indeces
		new(stoneChains_ + i) StoneChain(i);
	}

	currentChainState_   = chainStateStack_;
	nextStoneChain_      = stoneChains_;

	// set up initial chain state
	Point pt;
	for (pt.x = 0; pt.x < 11; pt.x++)
	{
		for (pt.y = 0; pt.y < 11; pt.y++)
		{
			if (pt.x == 0 || pt.x == 10 || pt.y == 0 || pt.y == 10)
			{
				currentChainState_->chainMap_[ BOARD_INDEX(pt) ] = &edgeChain_;
			}
			else
			{
				currentChainState_->chainMap_[ BOARD_INDEX(pt) ] = 0;
			}
		}
	}
	for (int i = 0; i < MAX_STONE_CHAINS; i++)
		currentChainState_->chainState_[i] = false;
}

void StoneChainManager::Clear()
{
	boardManager_.Clear();

	currentChainState_ = chainStateStack_;
	nextStoneChain_    = stoneChains_;
}

bool StoneChainManager::IsLegal(Move move)
{
	if (boardManager_.IsEmpty(move.point))
	{
		SetStone(move);		// try the move if the point is empty

		// if the move is a suicide or if it violates the ko rule, the move is illegal
		bool ret = !(boardManager_.IsCaptured(GetNewestStoneChain()) || boardManager_.IsKoViolated());
		Undo();
		return ret;
	}

	return false;
}

bool StoneChainManager::IsSafe(Move move)
{
	if (boardManager_.IsEmpty(move.point))
	{
		// remember the number of captured stone to see if the move just made is in a ko state
		int numPrisoners = boardManager_.GetNumPrisoners(move.color ^ 0x01);

		SetStone(move);		// try the move if the point is empty

		BitBoard9 libs;
		boardManager_.GetLiberties(GetNewestStoneChain(), &libs);

		// take the difference (if the difference is 1 and the lib of the new chain is 1, the chain is in ko state)
		numPrisoners = boardManager_.GetNumPrisoners(move.color ^ 0x01) - numPrisoners;
		bool isProfitableCapture = GetNewestStoneChain()->GetNumStones() <= numPrisoners;

		// if the move makes a chain that can be taken right away, or if it violates the ko rule, the move is not safe
		bool ret =  !((libs.CountStones() <= 1 && !isProfitableCapture) || boardManager_.IsKoViolated());
		Undo();
		return ret;
	}

	return false;
}

bool StoneChainManager::IsLegalForBothColor(Point pt)
{
	if (boardManager_.IsEmpty(pt))
	{
		Move move;
		move.point = pt;
		move.color = 0;

		SetStone(move);

		// if the move is a suicide or if it violates the ko rule, the move is illegal
		bool ret = !(boardManager_.IsCaptured(currentChainState_->chainMap_[ BOARD_INDEX(pt) ]) || boardManager_.IsKoViolated());
		Undo();

		move.color = 1;
		SetStone(move);

		// if the move is a suicide or if it violates the ko rule, the move is illegal
		ret = ret && !(boardManager_.IsCaptured(currentChainState_->chainMap_[ BOARD_INDEX(pt) ]) || boardManager_.IsKoViolated());
		Undo();

		return ret;
	}

	return false;
}

bool StoneChainManager::GetCapturingMove(StoneChain* target, BitBoard9* move)
{
	GetLiberties(target, move);

	return move->CountStones() == 1;
}

bool StoneChainManager::HasAtariChainAround(Move move)
{
	BitBoard9 dummy;
	StoneChain* chainN = currentChainState_->chainMap_[ BOARD_INDEX_NORTH(move.point) ];
	StoneChain* chainS = currentChainState_->chainMap_[ BOARD_INDEX_SOUTH(move.point) ];
	StoneChain* chainE = currentChainState_->chainMap_[ BOARD_INDEX_EAST (move.point) ];
	StoneChain* chainW = currentChainState_->chainMap_[ BOARD_INDEX_WEST (move.point) ];

	// move自体がアタリになっていなく、尚且つ周りのいずれかがあたりになっていれば。
	return !GetCapturingMove(currentChainState_->chainMap_[ BOARD_INDEX(move.point) ], &dummy) &&
		   ( (!boardManager_.IsEmpty(move.point.x, move.point.y - 1) && chainN->GetColor() != move.color && GetCapturingMove(chainN, &dummy)) ||
		     (!boardManager_.IsEmpty(move.point.x, move.point.y + 1) && chainS->GetColor() != move.color && GetCapturingMove(chainS, &dummy)) ||
		     (!boardManager_.IsEmpty(move.point.x + 1, move.point.y) && chainE->GetColor() != move.color && GetCapturingMove(chainE, &dummy)) ||
		     (!boardManager_.IsEmpty(move.point.x - 1, move.point.y) && chainW->GetColor() != move.color && GetCapturingMove(chainW, &dummy)) );
}  

bool StoneChainManager::HasBothColorStoneAround(Point pt)
{
	return boardManager_.HasBothColorStoneAround(pt);
}

void StoneChainManager::SetStone(Move move)
{
	boardManager_.SetStone(move);

	// copy the current chain state
	*(currentChainState_ + 1) = *(currentChainState_++);

	if (move.point.x != 0)	// if not pass
	{
		// init stone chain to merge surrounding chains
		nextStoneChain_->Initialize(move);

		// validate chain state
		currentChainState_->chainState_[nextStoneChain_->GetIndex()] = true;

		StoneChain* chain;
		if (!boardManager_.IsEmpty(move.point.x, move.point.y - 1))
		{
			chain = currentChainState_->chainMap_[ BOARD_INDEX_NORTH(move.point) ];
			if (chain->GetColor() == move.color)
			{
				nextStoneChain_->Amalgamate(chain);
				currentChainState_->chainState_[chain->GetIndex()] = false;
			}
			else
			{
				currentChainState_->chainState_[chain->GetIndex()] = !boardManager_.RemoveIfCaptured(chain);
			}
		}

		if (!boardManager_.IsEmpty(move.point.x, move.point.y + 1))
		{
			chain = currentChainState_->chainMap_[ BOARD_INDEX_SOUTH(move.point) ];
			if (chain->GetColor() == move.color)
			{
				nextStoneChain_->Amalgamate(chain);
				currentChainState_->chainState_[chain->GetIndex()] = false;
			}
			else
			{
				currentChainState_->chainState_[chain->GetIndex()] = !boardManager_.RemoveIfCaptured(chain);
			}
		}

		if (!boardManager_.IsEmpty(move.point.x - 1, move.point.y))
		{
			chain = currentChainState_->chainMap_[ BOARD_INDEX_WEST(move.point) ];
			if (chain->GetColor() == move.color)
			{
				nextStoneChain_->Amalgamate(chain);
				currentChainState_->chainState_[chain->GetIndex()] = false;
			}
			else
			{
				currentChainState_->chainState_[chain->GetIndex()] = !boardManager_.RemoveIfCaptured(chain);
			}
		}

		if (!boardManager_.IsEmpty(move.point.x + 1, move.point.y))
		{
			chain = currentChainState_->chainMap_[ BOARD_INDEX_EAST(move.point) ];	
			if (chain->GetColor() == move.color)
			{
				nextStoneChain_->Amalgamate(chain);
				currentChainState_->chainState_[chain->GetIndex()] = false;
			}
			else
			{
				currentChainState_->chainState_[chain->GetIndex()] = !boardManager_.RemoveIfCaptured(chain);
			}
		}

		// update the chain pointers
		BoardStonePointEnumerator stoneEnumerator;
		stoneEnumerator.Initialize(nextStoneChain_->GetStones());
		Point* libPt;

		while (libPt = stoneEnumerator.Next())
		{
			currentChainState_->chainMap_[ BOARD_INDEX_(libPt->x, libPt->y) ] = nextStoneChain_;
		}

		//BitBoard9 stones = nextStoneChain_->GetStones();
		//Point pt;
		//for (pt.x = 1; pt.x <= 9; pt.x++)
		//{
		//	for (pt.y = 1; pt.y <= 9; pt.y++)
		//	{
		//		if (!stones.IsEmpty(pt))
		//		{
		//			currentChainState_->chainMap_[ BOARD_INDEX(pt) ] = nextStoneChain_;
		//		}
		//	}
		//}
	}

	++nextStoneChain_;
}

void StoneChainManager::Undo()
{
	boardManager_.Undo();

	--currentChainState_;
	--nextStoneChain_;
}

#if defined (ENABLE_DEBUG_OUTPUT)
void StoneChainManager::ShowAllChains(char** comments)
{
	StoneChainEnumerator enumerator = GetEnumerator();
	StoneChain* chain;

	while (chain = enumerator.Next())
	{
		boardManager_.ShowChain(chain);

		printf("%s\n", comments[chain->GetIndex()]);
	}
}
#endif

StoneChainEnumerator StoneChainManager::GetEnumerator()
{
	enumerator_.Initialize(stoneChains_, nextStoneChain_, currentChainState_->chainState_);
	return enumerator_;
}

NeighborStoneChainEnumerator StoneChainManager::GetNeighborEnumeratorOf(StoneChain* chain)
{
	BitBoard9 surrounding;
	boardManager_.GetSurroundingStones(chain, &surrounding);

	NeighborStoneChainEnumerator enumerator;
	enumerator.Initialize(surrounding, currentChainState_->chainMap_);

	return enumerator;
}

NeighborStoneChainEnumerator StoneChainManager::GetNeighborEnumeratorAround(Point pt)
{
	BitBoard9 surrounding;
	boardManager_.GetSurroundingStones(pt, &surrounding);

	NeighborStoneChainEnumerator enumerator;
	enumerator.Initialize(surrounding, currentChainState_->chainMap_);

	return enumerator;
}

StoneChain* StoneChainManager::GetChainAt(Point pt)
{
	return IsEmpty(pt) ? 0 : currentChainState_->chainMap_[BOARD_INDEX(pt)];
}